import click
from . import entities, generate, openapi


@click.command()
@click.option(
    "--filename",
    help="OpenAPI filename (JSON file)",
    default="apidocs/openapi30.json",
)
def generate_routes(filename):
    """Create `routes.py` and `test_routes.py` based on OpenAPI file.

    :param filename: json OpenAPI filename
    :type filename: str
    :param package_name: package name
    :type package_name: str
    """
    file = openapi.load_json_file(filename)
    try:
        file["info"]["x-url-base"]
        package_name = file["info"]["x-package-name"]
    except KeyError:
        print(
            "Add 'x-package-name' and 'x-base-url' to the info block of 'openapi30.json' file "
        )
        return
    routes = openapi.parse_routes(oa_file=file, package_name=package_name)
    generate.routes(routes, package_name)
    generate.test_routes(routes, package_name)
    generate.apidocs(package_name=package_name)


@click.command()
@click.option("--package-name", help="Name of package to create views for.")
@click.option(
    "--filename", help="OpenAPI filename (JSON file)", default="api-spec.json"
)
def generate_views(filename: str, package_name: str):
    """Generate `views.py` with handlers based on OpenAPI file.

    :param filename: json OpenAPI filename
    :type filename: str
    :param package_name: package name
    :type package_name: str
    """
    file = openapi.load_json_file(filename)
    routes = openapi.parse_routes(oa_file=file, package_name=package_name)
    generate.views(routes, package_name)


@click.command()
@click.option(
    "--filename",
    help="OpenAPI Entities filename (JSON file)",
    default="apidocs/generated_entities.json",
)
@click.option(
    "--entitymod",
    help="Entity module to get the entities from (e.g.: zsnl_domains.case_management.entities)",
    required=True,
)
def generate_entities(filename: str, entitymod: str):
    """Generate json with handlers based on OpenAPI file.

    Saves a list of generated entities to the given json formatted file

    Args:
        filename (str): location to save output to
        entitymod (str): Name of module to get entities from
    """
    entity_generator = entities.EntityGenerator(entity_module_name=entitymod)

    with open(filename, "w", encoding="utf-8") as file:
        file.write(entity_generator.generate_json() + "\n")



import codecs
import json
import logging
import time
from collections import defaultdict, namedtuple
from functools import wraps
from minty import Base
from pyramid.httpexceptions import HTTPForbidden, HTTPUnauthorized
from redis import StrictRedis


class SessionRetrieverError(Exception):
    """Base for session retrieval related exceptions."""

    pass


class SessionNotFoundError(SessionRetrieverError, ValueError):
    """Session can't be found in Redis."""

    pass


class SessionDataNotFoundError(SessionRetrieverError, ValueError):
    """Session data can't be found in Redis."""

    pass


class SessionRetriever(Base):
    """HTTP session retriever.

    retrieves a session stored by the Perl application from the Redis store
    """

    __slots__ = ["redis", "decoder"]

    def __init__(self, redis: StrictRedis):
        """Initialize a HTTP session retriever.

        :param redis: Redis connection object to retrieve session data
        :type redis: StrictRedis
        """
        self.redis = redis

    def retrieve(self, session_id: str) -> dict:
        """Retrieve a session from Redis.

        :param session_id: session id to retrieve
        :type session_id: str
        :raises SessionNotFoundError: if the session cannot be found
        :raises SessionDataNotFoundError: if the session is found, but contains
                                          no data
        :return: the session data
        :rtype: dict
        """
        now = time.time()
        with self.statsd.get_timer("retrieve").time("total_time"):
            with self.statsd.get_timer("retrieve").time(
                "redis.get_expires.time"
            ):
                expiration = self.redis.get(f"expires:{session_id}")

            if expiration is None:
                self.logger.info(f"Session '{session_id}' not found")
                self.statsd.get_counter("retrieve").increment(
                    "redis.get_session.not_found.count"
                )
                raise SessionNotFoundError(session_id)

            if float(expiration) < now:
                self.logger.info(f"Session '{session_id}' has expired")
                self.statsd.get_counter("retrieve").increment(
                    "redis.get_session.not_found.count"
                )
                raise SessionNotFoundError(session_id)

            with self.statsd.get_timer("retrieve").time(
                "redis.get_session.time"
            ):
                session_data_raw = self.redis.get(f"json:session:{session_id}")

            if session_data_raw is None:
                self.logger.info(f"Session '{session_id}' not found")
                self.statsd.get_counter("retrieve").increment(
                    "redis.get_session.not_found.count"
                )
                raise SessionDataNotFoundError(session_id)

            session_data_decoded = codecs.decode(session_data_raw, "base64")
            session_data = json.loads(session_data_decoded)

            self.statsd.get_counter("retrieve").increment(
                "redis.get_session.count"
            )
            self.logger.info(f"Session '{session_id}' retrieved")

            return session_data


def redis_from_config(config: dict) -> StrictRedis:
    """Create a Redis cleint from variables in the config parameter.

    :param config: config variables
    :type config: dict
    :return: configured redis client
    :rtype: StrictRedis
    """
    redis_conf = config["redis"]["session"]

    redis = StrictRedis(**redis_conf)
    return redis


def session_manager_factory(infra_factory):
    """Create a configured SessionRetriever class.

    :param infra_factory: infrastructure factory class
    :type infra_factory: InfrastructureFactory
    :return: configured SessionRetriever class to retrieve session information
        from session store
    :rtype: SessionRetriever
    """
    infra_factory.register_infrastructure(
        name="redis", infrastructure=redis_from_config
    )
    redis = infra_factory.get_infrastructure(
        context=None, infrastructure_name="redis"
    )
    return SessionRetriever(redis)


UserInfo = namedtuple("UserInfo", "user_uuid permissions")


def get_logged_in_user(request, logger):
    """ Get the logged in user from the given pyramid request object

    Args:
        request: the pyramid request object to get the user from
        logger: the logging.getLogger() instance to log to

    Returns:
        user_info: dict containing values from handle_session_data()
    """

    try:
        request.assert_platform_key()

        user_uuid = request.get_platform_user_uuid()
        permissions = defaultdict(lambda: True, pip_user=False)

        user_info = UserInfo(user_uuid, permissions)
        logger.info("Platform key used successfully")
    except (AttributeError, KeyError):
        # If it's not specified, back to regular scheduled programming
        # (session check)
        try:
            session_data = request.retrieve_session()
        except (SessionRetrieverError, KeyError) as e:
            logger.info(f"Invalid session: '{e}'")
            raise HTTPUnauthorized(
                content_type="application/json",
                json_body={"error": "Unauthorized"},
            )

        user_info = handle_session_data(
            session_data=session_data, logger=logger
        )

    return user_info


def check_user_permission(permissions, user_info):
    """ Returns the first permission the user validates to.

    Looks up every permission in the given list of permissions, and when one
    validates, it returns that perm as a result

    Args:
        permissions: List containing permissions in string format
        user_info: the result from handle_session_data() to check against

    Returns:
        perm: the string permission we first positively validated against
    """

    # User permissions will be set on login, based on a table of roles
    # and access rights.
    #
    # This generally means someone with "admin" rights will also have
    # the other permissions set, for instance.
    for permission in permissions:
        try:
            perm = user_info.permissions[permission]
        except KeyError:
            perm = False

        if perm:
            break

    return perm


def protected_route(permission="gebruiker", *args):
    """Check session and inject `UserInfo` in wrapped view .

    zaaksysteem_session cookie is checked for validity and `UserInfo` is
    injected in the view.

    :param view: view function
    :type view: pyramid view
    :raises HTTPUnauthorized: session not valid or non-existant
    :return: view with `UserInfo` injected
    :rtype: wrapped view
    """
    permission_list = [permission, *args]

    def protected_view_wrapper(view):
        @wraps(view)
        def view_wrapper(request):
            logger = logging.getLogger(view.__name__)

            user_info = get_logged_in_user(request, logger)
            perm = check_user_permission(permission_list, user_info)

            if perm:
                return view(request=request, user_info=user_info)

            raise HTTPForbidden(
                json={"error": "You do not have permission to access this."}
            )

        return view_wrapper

    return protected_view_wrapper


def handle_session_data(session_data: dict, logger) -> UserInfo:
    """Handle session data for different types of users.

    Current users: zaaksyteem user & pip_user

    :param session_data: session
    :type session_data: dict
    :param logger: logger
    :type logger: Logger
    :raises HTTPUnauthorized: if user is not Authorized
    :return: user_info object
    :rtype: UserInfo
    """
    try:
        user_info_raw = session_data.get("__user", False)
        if user_info_raw:
            user_info_decoded = json.loads(user_info_raw)
            user_info = UserInfo(
                user_uuid=user_info_decoded["subject_uuid"],
                permissions=user_info_decoded["permissions"],
            )
            return user_info

        user_info_pip = session_data.get("pip", False)
        if user_info_pip:
            user_info = UserInfo(
                user_uuid=user_info_pip["user_uuid"],
                permissions={"pip_user": True},
            )
            return user_info
    except KeyError as e:
        logger.info(f"Unauthorized: error in session data: '{e}'")
        raise HTTPUnauthorized(
            content_type="application/json",
            json_body={"error": "Unauthorized"},
        )
    except TypeError as e:
        logger.info(f"Unauthorized: error in session data: '{e}'")
        raise HTTPUnauthorized(
            content_type="application/json",
            json_body={"error": "Unauthorized"},
        )

    raise HTTPUnauthorized(
        content_type="application/json", json_body={"error": "Unauthorized"}
    )



import logging
import re
from minty_pyramid.session_manager import (
    check_user_permission,
    get_logged_in_user,
)
from pyramid.httpexceptions import HTTPForbidden

log = logging.getLogger(__name__)


# This will be moved to minty-pyramid in a next iteration
class JSONAPIEntityView:
    """View Base class for Pyramid Views which return Entity objects

    Use this as a base class for your own view. E.g.

        class DogView(JSONAPIEntityView):
            view_mapper = {
                "GET": {
                    "get_dog": {
                        "cq": "query",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "get_dog_by_uuid",
                        "from": {"request_params": {"uuid": "uuid"}},
                    },
                    "filter_dogs": {
                        "cq": "query",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "get_dog_by_uuid",
                        "from": {"request_params": {"filter[]": "filter_params"}},
                    }
                },
                "POST": {
                    "create_dog": {
                        "cq": "command",
                        "auth_permissions": ["gebruiker"],
                        "domain": "zsnl_domains.case_management",
                        "run": "create_dog",
                        "from": {
                            "request_params": {"uuid": "uuid"},
                            "json": {"_all": True},
                        },
                        "command_primary": "uuid",
                        "entity_type": "dog",
                    }
                }
            }

            def create_link_from_entity(
                self, entity=None, entity_type=None, entity_id=None
            ):
                if entity is not None:
                    entity_type = entity.entity_type
                    entity_id = str(entity.entity_id)

                if entity_type == "dog":
                    return self.request.route_url(
                        "get_dog_object", _query={"uuid": entity_id}, _scheme="https"
                    )

    The view_mappers has as it's first key the method of the API call, e.g.
    POST or GET. Define the action in the url as the first key of the following
    dict (get_dog, create_dog) in the above example.

    Other parameters:
        cq (Enum("query", "command")):  Defines whether to call on the Query or
                                        Command class
        auth_permissions (List[str]): The list of permissions to check
        domain (str): The domain to run this command or query against
        run (str): the command or query to run
        from (dict): A dict of parameters, see below
        command_primary (str):  The "reference" of the command extracted from
                                the given "parameter
        entity_type (str):  The string entity_type, so we know when we got no
                            Entity as a return value, which type to set

    The "from" parameter extracts input parameters from either the given
    "request_parameters" or the "json_body" of the request. The following
    dict contains a key value pair. The key correspondents to the key from the
    input. The value correspondents to the key for the command to run.

    Using array[] options in request_params:
    When a key of a request_params ends with the characters [] (e.g.: see
    filter_dogs above), then it will transform the variables into a dict.

    For instance: ?filter[name]=frits&filter[gender]=male
    Will transform into:
        {
            "name": "frits",
            "gender": "male"
        }

    """

    def __init__(self, context, request):
        """Initializes this object from Pyramid"""

        self.request = request
        self.call = request.matched_route.name

    def _check_protected_route(self, calldata):
        """Verifies authentication permissions against our session"""

        user_info = get_logged_in_user(self.request, log)
        if not check_user_permission(calldata["auth_permissions"], user_info):
            raise HTTPForbidden(
                json={"error": "You do not have permission to access this."}
            )

        return user_info

    def _get_cqrs_instance(self, calldata, user_info):
        """Gets the command or query instance for the called api view"""
        cq = calldata["cq"]

        cmdqry = getattr(self.request, f"get_{cq}_instance")
        return cmdqry(
            domain=calldata["domain"],
            user_info=user_info,
            user_uuid=user_info.user_uuid,
        )

    def __call__(self):
        """Reads and runs the called API route defined in view_mapper"""

        calldata = self.view_mapper[self.request.method][self.call]

        user_info = self._check_protected_route(calldata)
        cqinstance = self._get_cqrs_instance(calldata, user_info)

        run = getattr(cqinstance, calldata["run"])
        params = self._collect_params(calldata)

        response = None
        if "wrapper" in calldata:
            wrapper = getattr(self, calldata["wrapper"])
            response = wrapper(run, params)
        else:
            response = run(**params)

        return self.transform_response_from_command_or_query(
            response, calldata, params
        )

    def transform_response_from_command_or_query(
        self, response, calldata, params
    ):
        """Transforms a response from a command to a proper JSONApi structure

        Args:
            response (Optional[Entity]): The response from our query or command
            calldata (dict): The configuration defined in our view_mapper
            params (dict): The collected parameters from our query and jsonbody

        Returns:
            Structured dict for our jsonapi serializer:

                {
                    "data": {
                        "links": {
                            "self": "http://url",
                        },
                        "type": "name_of_entity",
                        "id": "12434-abcde-1234-acbewe-233aba"
                    }
                }

        """
        if calldata["cq"] == "command":
            try:
                if (
                    calldata["command_primary"]
                    and params[calldata["command_primary"]]
                ):
                    return {
                        "data": {
                            "links": {
                                "self": self.create_link_from_entity(
                                    entity_type=calldata["entity_type"],
                                    entity_id=params[
                                        calldata["command_primary"]
                                    ],
                                )
                            },
                            "type": calldata["entity_type"],
                            "id": params[calldata["command_primary"]],
                        }
                    }
            except KeyError:
                return {"data": None}

        if calldata["cq"] == "query":
            if response:
                return {
                    "data": response,
                    "links": {"self": self.request.current_route_path()},
                }
            else:
                return {"data": None}

    def _collect_params_as_array(self, k, v, params):
        for pk, pv in self.request.params.items():
            match = re.search(k + r"\[(.+)\]", pk)

            if match:
                if k not in params:
                    params[v] = {}

                params[v][match[1]] = pv

    def _collect_params(self, calldata):
        """Collects parameters according to the configuration in view_mapper"""
        params = {}

        if "request_params" in calldata["from"]:
            for k, v in calldata["from"]["request_params"].items():
                # Special array functionality
                if k.endswith("[]"):
                    k = k.replace("[]", "")
                    self._collect_params_as_array(k, v, params)

                    continue

                if k not in self.request.params:
                    continue

                params[v] = self.request.params[k]

        if "json" in calldata["from"]:
            if (
                "_all" in calldata["from"]["json"]
                and calldata["from"]["json"]["_all"] is True
            ):
                for k in self.request.json_body.keys():
                    params[k] = self.request.json_body[k]
            else:
                for k, v in calldata["from"]["json"].items():
                    if k not in self.request.json_body:
                        continue
                    params[v] = self.request.json_body[k]

        return params



import datetime
from minty.entity import Entity, EntityCollection
from pyramid.renderers import JSON
from uuid import UUID


class jsonapi(JSON):
    jsonapi_adapters = {
        Entity: "_entity_adapter",
        EntityCollection: "_entity_collection_adapter",
        UUID: "_uuid_adapter",
        datetime.datetime: "_datetime_adapter",
    }

    def __init__(self, *args, **kwargs):
        """ Constructs the jsonapi class based on Pyramid JSON implementation.

        This will add some default adapters to the system, like UUID, entity
        parsing and datetime handling
        """
        super().__init__(*args, **kwargs)

        for obj, adapter in self.jsonapi_adapters.items():
            self.add_adapter(obj, getattr(self, adapter))

    def __call__(self, info):
        """ Returns a plain JSON-encoded string with content-type
        ``application/json``. The content-type may be overridden by
        setting ``request.response.content_type``."""

        # We had top copy this from the Pyramid JSON implementation, because
        # we would like to inject the view on self, so we can use the view
        # class in adapters for generating urls for example
        def _render(value, system):
            request = system.get("request")
            self.view = system.get("view")

            if request is not None:
                response = request.response
                ct = response.content_type
                if ct == response.default_content_type:
                    response.content_type = "application/json"
            default = self._make_default(request)
            return self.serializer(value, default=default, **self.kw)

        return _render

    def _entity_collection_adapter(self, obj, request):
        return [entity for entity in obj]

    def __links_for_entity(self, entity):
        self_link = self.view.create_link_from_entity(entity)
        if self_link:
            return {"links": {"self": self_link}}
        else:
            return {}

    def _entity_adapter(self, obj, request):
        """ Transforms an entity in a JSON:API styled format

        See renderers documentation of Pyramid for more information about
        using adapters
        """

        object_attributes = obj.entity_dict()
        object_relationships = {}

        # Retrieve relationships from entity
        for relationshipkey in obj.entity_relationships:
            relationship = getattr(obj, relationshipkey)
            if relationship is not None:
                list_of_relationships = False
                is_list = None
                if isinstance(relationship, list):
                    list_of_relationships = relationship
                    is_list = True
                else:
                    list_of_relationships = [relationship]

                for relationshipcontent in list_of_relationships:
                    relationshipdata = relationshipcontent.dict()

                    metadata = {}
                    if relationshipdata["entity_meta_summary"] is not None:
                        metadata["meta"] = {
                            "summary": relationshipdata["entity_meta_summary"]
                        }

                    relationship_response = {
                        "data": {
                            "type": relationshipdata["entity_type"],
                            "id": relationshipdata["entity_id"],
                        },
                        **metadata,
                        **self.__links_for_entity(relationshipcontent),
                    }

                    if not is_list:
                        object_relationships[
                            relationshipkey
                        ] = relationship_response
                    else:
                        if relationshipkey not in object_relationships:
                            object_relationships[relationshipkey] = []

                        object_relationships[relationshipkey].append(
                            relationship_response
                        )

            del object_attributes[relationshipkey]

        for entity_id__field in obj.entity_id__fields:
            del object_attributes[entity_id__field]

        relationships_block = {}
        if object_relationships != {}:
            relationships_block["relationships"] = object_relationships

        return {
            "type": obj.entity_type,
            "id": obj.entity_id,
            "meta": {"summary": obj.entity_meta_summary},
            "attributes": object_attributes,
            **relationships_block,
            **self.__links_for_entity(obj),
        }

    def _uuid_adapter(self, obj, request):
        """ Transforms a UUID in a JSONApi styled format

        See renderers documentation of Pyramid for more information about
        using adapters
        """
        return str(obj)

    def _datetime_adapter(self, obj, request):
        """ Transforms a datetime object in a JSONApi styled format

        See renderers documentation of Pyramid for more information about
        using adapters
        """
        return obj.isoformat()



"""Entities generator for JSONAPI style responses in our Pyramid framework

This module generates a openapi 3.0 compatible entity file containing a list
of components generated from the given zsnl_domains entity module.

In short: it uses pydantic.schema() for every entity this module can find in
the given domain, and turns it into a JSONAPI compatible output for your
openapi specification

Example:

    $ generate-entities --entitymod=zsnl_domains.case_management.entities

"""

import json
import jsonpath_ng as jp
from copy import deepcopy
from importlib import import_module
from minty.entity import Entity
from typing import List


class EntityGenerator:
    """The actual generator logic to transform pydantic entities to jsonapi

    An instance of this object turns the given entities in the given entities
    module into a JSONAPI and OpenAPI compatible "components" output.
    """

    entitymod: object = None
    entities: List = []

    def __init__(self, entity_module_name):
        """Initializes an instance of this entity generator

        Args:
            entity_module_name (module): name of the module containing all the
                entities in the __all__ attribute.
        """
        self._load_entity_module(entity_module_name)
        self._load_entities()

    def _load_entity_module(self, entity_module_name):
        """Imports the entity module given by name

        Args:
            entity_module_name (module): name of the module containing all the
                entities in the __all__ attribute.
        """
        entities_module = import_module(entity_module_name)
        self.entitymod = entities_module

    def _load_entities(self):
        """Adds all class instances of the loaded entity module

        Appends every entity class instance in attribute self.entities
        """
        for entity in self.entitymod.__all__:
            self.entities.append(getattr(self.entitymod, entity))

    def generate(self):
        """Generates a JSONAPI openapi compatible schame output

        For every entity in self.entities it generates a JSONAPI compatible
        entry in an openapi compatible json format
        """
        jsonapischema = {}
        definitionsmap = {"entities": {}}
        for entity in self.entities:
            if issubclass(entity, Entity):
                self._transform_entity(entity, jsonapischema, definitionsmap)

        for name, definition in definitionsmap.items():
            jsonapischema[name] = definition

        # Make sure every ref link gets updated to the proper location
        matches = jp.parse("$..'$ref'").find(jsonapischema)
        for match in matches:
            if match.value.startswith("Entity"):
                refvalue = match.value.replace(
                    "#/definitions/", "#/components/schemas/"
                )
            else:
                refvalue = match.value.replace(
                    "#/definitions/", "#/components/schemas/Def"
                )

            match.context.value["$ref"] = refvalue

        return {"components": {"schemas": jsonapischema}}

    def generate_json(self):
        """Generates a string json response

        Same as generate, but instead of returning a python dict, it will
        return a json formatted string of it.
        """
        return json.dumps(self.generate(), sort_keys=True, indent=2)

    def generate_json_api_object(
        self,
        meta_schema,
        attribute_schema,
        entity_type,
        relationships={},
        schema_required=[],
    ):
        """Generate a jsonapi formatted python object

        Will generate a json_api_object from the given arguments. Including
        links, attributes, relationships.

        Args:
            meta_schema (dict): A schema dictionary containing meta fields
            attribute_schema (dict): A schema dictionary containing fields for
                the attribute part of jsonapi
            entity_type (str): The name of this object (type), e.g.: "case"
            relationships (dict): An instruction of a relationship (key) and
                information about the relationship (is_list) and it's
                (entity_type)
            schema_required (list): A list of required parameters from original
                schema
        """
        relationship_schema = {
            key: self._generate_json_api_relationship(key, value)
            for key, value in relationships.items()
        }

        required = [key for key in schema_required if key in attribute_schema]

        return {
            "properties": {
                "type": {"type": "string", "example": entity_type},
                "id": {"type": "string", "format": "uuid"},
                "meta": {"type": "object", "properties": meta_schema},
                "attributes": {
                    "type": "object",
                    "properties": attribute_schema,
                    "required": required,
                },
                "relationships": {
                    "type": "object",
                    "properties": relationship_schema,
                },
                "links": {
                    "type": "object",
                    "properties": {
                        "self": {
                            "type": "string",
                            "format": "uri",
                            "example": "/api/v2/get_something?uuid=1a2b3c4d-1a2b-1a2b-1a2b-1a2b3c4d5e6f",
                        }
                    },
                },
            },
            "required": [
                "type",
                "id",
                "meta",
                "attributes",
                "relationships",
                "links",
            ],
        }

    def _generate_json_api_relationship(self, name, relationship_info):
        """Generate the jsonapi formatted part for the relationships

        Args:
            name (str): Name of the relationship, e.g: "brand"
            relationship_info (dict): An instruction of the relationship with
                information about the relationship (is_list) and it's
                (entity_type)
        """
        relationship_type = {"type": "string"}

        if relationship_info["entity_type"]:
            relationship_type["example"] = relationship_info["entity_type"]

        relationship_schema_output = {
            "type": "object",
            "properties": {
                "links": {
                    "type": "object",
                    "properties": {
                        "self": {
                            "type": "string",
                            "format": "uri",
                            "example": "/api/v2/get_something?uuid=1a2b3c4d-1a2b-1a2b-1a2b-1a2b3c4d5e6f",
                        }
                    },
                },
                "data": {
                    "type": "object",
                    "properties": {
                        "type": relationship_type,
                        "id": {"type": "string", "format": "uuid"},
                    },
                },
            },
        }

        if relationship_info["is_list"]:
            relationship_schema_output = {
                "type": "array",
                "items": relationship_schema_output,
            }

        return relationship_schema_output

    def _transform_entity_to_jsonapi(self, entity, schema):
        """Transforms a given entity in a jsonapi formatted object

        Transforms the given entity and schema into a jsonapi compatible object

        Args:
            entity (:obj:`Entity`): The pydantic based entity to convert
            schema (dict): The schema generated from the entity (entity.schema())
        """
        output_schema = deepcopy(schema)

        jsonschema = self.generate_json_api_object(
            meta_schema=self._get_meta_from_entity(
                entity, schema["properties"]
            ),
            attribute_schema=self._get_attributes_from_entity(
                entity, schema["properties"]
            ),
            relationships=self._get_relationships_from_entity(
                entity, schema["properties"]
            ),
            entity_type=entity.__fields__.get("entity_type").default,
            schema_required=output_schema["required"],
        )

        output_schema["properties"] = jsonschema["properties"]
        if "required" in output_schema:
            del output_schema["required"]
        if "definitions" in output_schema:
            del output_schema["definitions"]

        del output_schema["title"]

        return output_schema

    def _get_meta_from_entity(self, entity, properties):
        """Generate the jsonapi formatted part for the meta part

        Args:
            entity (:obj:`Entity`): The pydantic based entity to convert
            properties (dict): The properties part of the entity
        """
        meta_properties = {}
        for prop in properties.keys():
            if (
                prop.startswith("entity_meta_")
                and prop != "entity_meta__fields"
            ):
                meta_properties[prop.replace("entity_meta_", "")] = properties[
                    prop
                ]

        return meta_properties

    def _get_attributes_from_entity(self, entity, properties):
        """Generate the jsonapi formatted part for the attributes part

        Args:
            entity (:obj:`Entity`): The pydantic based entity to convert
            properties (dict): The properties part of the entity
        """
        attribute_properties = deepcopy(properties)
        for prop in properties.keys():
            if (
                prop.startswith("entity_")
                or prop in entity.__fields__.get("entity_id__fields").default
                or prop
                in entity.__fields__.get("entity_relationships").default
            ):
                del attribute_properties[prop]

        return attribute_properties

    def _get_relationships_from_entity(self, entity, properties):
        """Generate the jsonapi formatted part for the relationships part

        Args:
            entity (:obj:`Entity`): The pydantic based entity to convert
            properties (dict): The properties part of the entity
        """
        relationships = {}
        for relationship in entity.__fields__.get(
            "entity_relationships"
        ).default:
            is_list = False
            if (
                "type" in properties[relationship]
                and properties[relationship]["type"] == "array"
            ):
                is_list = True

            relationship_type = None
            try:
                relationship_type = (
                    entity.__fields__.get(relationship)
                    .type_.__fields__.get("entity_type")
                    .default
                )
            except Exception:
                pass

            relationships[relationship] = {
                "is_list": is_list,
                "entity_type": None,
            }

            if relationship_type:
                relationships[relationship]["entity_type"] = relationship_type

        return relationships

    def generate_jsonapi_envelope_for(self, schema, multiple=False):
        """Generate an envelope for the jsonapi formatted object

        The envelope contains, depending of the context (list or a single
        result) the data and links attributes

        Args:
            schema (dict): The schema to include in the envelope
            multiple (bool): Indicates whether we want the response in a single
                jsonapi formatted response, or in a list format
        """
        envelope = {
            "type": "object",
            "properties": {
                "data": None,
                "links": {
                    "type": "object",
                    "properties": {
                        "self": {
                            "type": "string",
                            "format": "uri",
                            "example": "/api/v2/get_something?uuid=1a2b3c4d-1a2b-1a2b-1a2b-1a2b3c4d5e6f",
                        }
                    },
                },
            },
        }

        if multiple:
            envelope["properties"]["data"] = {"type": "array", "items": schema}
        else:
            envelope["properties"]["data"] = schema

        return envelope

    def _transform_entity(self, entity, jsonapischema, definitionsmap):
        """Transforms the given entity in a jsonapi formatted object

        Args:
            entity (:obj:`Entity`): The pydantic based entity to convert
            jsonapischema (dict): The schema to transform
            definitionsmap (dict): A dictorary of other objects to transform
        """
        schema = entity.schema()
        # load definitions
        if "definitions" in schema:
            for name, definition in deepcopy(schema["definitions"]).items():
                newname = f"Def{name}"
                if "entity_id__fields" in definition["properties"]:
                    # Is an entity, skip over it
                    definitionsmap["entities"][name] = f"Entity{name}"
                    continue

                definition["title"] = newname
                definitionsmap[newname] = definition

        single = self.generate_jsonapi_envelope_for(
            self._transform_entity_to_jsonapi(entity, schema)
        )
        multiple = self.generate_jsonapi_envelope_for(
            self._transform_entity_to_jsonapi(entity, schema), multiple=True
        )

        jsonapischema[f"Entity{entity.__name__}"] = single
        jsonapischema[f"Entity{entity.__name__}List"] = multiple



import json
from functools import partial


def parse_routes(oa_file, package_name: str) -> list:
    """Parse OpenAPI file and return a list of routes.

    This function expects the following directory structure in the application:
    `{package_name}/views/{x-view}`
    where `x-view` is the filename for the handlers specified in the OpenAPI
    file.

    :param oa_file: OpenAPI dict
    :type oa_file: dict
    :param package_name: package name
    :type package_name: str
    :return: list of routes
    :rtype: list
    """
    routes = []
    urls = oa_file["info"]["x-url-base"]
    for url in urls:
        r = {
            "route": url + "openapi/{filename}",
            "handler": "apidocs_fileserver",
            "renderer": "json",
            "method": "GET",
            "view": f"{package_name}.views.apidocs.apidocs_fileserver",
            "handler_file": "apidocs",
            "params": [],
            "grouped_params": {},
        }
        routes.append(r)

    for route in oa_file["paths"]:
        r = oa_file["paths"][route]
        for method in r:
            h_file = r[method]["x-view"]
            handler = r[method]["operationId"]
            try:
                params = r[method]["parameters"]
            except KeyError:
                params = []

            viewclass = None
            if "x-view-class" in r[method].keys():
                viewclass = r[method]["x-view-class"]

            out = {
                "route": route,
                "method": method.upper(),
                "handler": handler,
                "renderer": "json",
                "view": f"{package_name}.views.{h_file}.{handler}",
                "handler_file": h_file,
                "params": params,
                "grouped_params": group_params_by_query_method(
                    r[method]["parameters"] if params else {}
                ),
            }

            if viewclass is not None:
                out["view"] = f"{package_name}.views.{h_file}.{viewclass}"
                out["viewclass"] = viewclass

            routes.append(out)
    return routes


def load_json_file(filename: str) -> dict:
    """Load an OpenAPI JSON file.

    :param filename: filename
    :type filename: str
    :return: OpenAPI file
    :rtype: dict
    """
    with open(filename, "r", encoding="utf-8") as jsonfile:
        return json.load(jsonfile)


def group_params_by_query_method(params: list) -> dict:
    """Group parameters in openapi file by query type.

    The grouping ensures that each parameter is fetched from the url path or query
    string as it is defined in the openapi spec.
    Types: "path", "query_required", "query_optional", "query_multiple"

    :param params: parameters
    :type params: list
    :return: grouped parameters
    :rtype: dict
    """
    param_types = {}
    for param in params:
        add_to_type_group = partial(
            add_object_to_grouping, item=param, grouping=param_types
        )
        if param["in"] == "path":
            param_types = add_to_type_group(group_name="path")
        elif param["in"] == "query" and param["schema"]["type"] == "array":
            param_types = add_to_type_group(group_name="query_multiple")
        elif param["in"] == "query" and param["required"] is True:
            param_types = add_to_type_group(group_name="query_required")
        elif param["in"] == "query" and param["required"] is False:
            param_types = add_to_type_group(group_name="query_optional")

    return param_types


def add_object_to_grouping(
    item: object, grouping: dict, group_name: str
) -> dict:
    """Add item to grouping with group name.

    :param item: item to add to grouping
    :type item: object
    :param grouping: previously grouped items
    :type grouping: dict
    :param group_name: group to append to
    :type group_name: str
    :return: grouped items
    :rtype: dict
    """
    group = grouping.get(group_name, [])
    group.append(item)
    grouping[group_name] = group
    return grouping
